from __future__ import annotations

import copy
import logging
import socket
import sys
from collections.abc import Callable
from types import TracebackType
from typing import Any, Optional, Union

import jsonpickle
import requests

from raygun4py import raygunmsgs, utilities

DEFAULT_CONFIG: dict[str, Any] = {
    "before_send_callback": None,
    "grouping_key_callback": None,
    "filtered_keys": [],
    "ignored_exceptions": [],
    "proxy": None,
    "transmit_global_variables": True,
    "transmit_local_variables": True,
    "enforce_payload_size_limit": True,
    "log_payload_size_limit_breaches": True,
    "transmit_environment_variables": True,
    "userversion": "Not defined",
    "user": None,
    "http_timeout": 10.0,
}


# Type aliases for clarity
UserInfo = Optional[Union[dict[str, Any], str]]
HttpRequest = Optional[dict[str, Any]]
Tags = Optional[list[str]]
CustomData = Optional[dict[str, Any]]
ExcInfo = tuple[type[BaseException], BaseException, Optional[TracebackType]]
BeforeSendCallback = Callable[[dict[str, Any]], Optional[dict[str, Any]]]
GroupingKeyCallback = Callable[[raygunmsgs.RaygunMessage], str]


class RaygunSender:
    """
    A sender for reporting errors to Raygun.

    Attributes:
        api_key (str): The API key for Raygun.
    """

    log: logging.Logger = logging.getLogger(__name__)

    api_key: str | None = None
    endpointprotocol: str = "https://"
    endpointhost: str = "api.raygun.io"
    endpointpath: str = "/entries"
    process_tags: list[str] = []
    process_custom_data: dict[str, Any] = dict()

    # Config attributes set dynamically in __init__
    before_send_callback: BeforeSendCallback | None
    grouping_key_callback: GroupingKeyCallback | None
    filtered_keys: list[str]
    ignored_exceptions: list[type[Exception]]
    proxy: dict[str, Any] | None
    transmit_global_variables: bool
    transmit_local_variables: bool
    enforce_payload_size_limit: bool
    log_payload_size_limit_breaches: bool
    transmit_environment_variables: bool
    userversion: str
    user: UserInfo
    http_timeout: float

    def __init__(
        self, api_key: str | None, config: dict[str, Any] | None = None
    ) -> None:
        """
        Initialize a RaygunSender.

        Parameters:
            api_key (str): The API key for Raygun.
            config (dict, optional): Configuration options. Defaults to an empty dictionary.
        """
        if api_key:
            if not isinstance(api_key, str):
                self.log.error(
                    f"Expected api_key of type str, but got {type(api_key).__name__}. Errors will not be transmitted."
                )
                self.api_key = None
            else:
                self.api_key = api_key
        else:
            self.log.warning(
                "RaygunProvider error: api_key not set, errors will not be transmitted."
            )

        try:
            import ssl  # noqa: F401
        except ImportError:
            self.log.warning(
                "RaygunProvider error: No SSL support available, cannot send. Please"
                "compile the socket module with SSL support."
            )

        # Set up the default values
        default_config = utilities.snakecase_dict(copy.deepcopy(DEFAULT_CONFIG))
        default_config.update(utilities.snakecase_dict(config or {}))
        for k, v in default_config.items():
            setattr(self, k, v)

    def set_version(self, version: str) -> None:
        """
        Set the version for the error reports.

        Parameters:
            version (str): Version string.
        """
        if isinstance(version, str):
            self.userversion = version

    def set_user(self, user: UserInfo) -> None:
        """
        Set user information for the error reports.

        Parameters:
            user: User information.
        """
        self.user = user

    def set_tags(self, tags: list[str]) -> None:
        """
        Set tags for the error reports.

        Parameters:
            tags (list): List of tags.
        """
        if type(tags) is list:
            self.process_tags = tags

    def set_customdata(self, custom_data: dict[str, Any]) -> None:
        if type(custom_data) is dict:
            self.process_custom_data = custom_data

    def ignore_exceptions(self, exceptions: list[type[Exception]]) -> None:
        """
        Set exceptions that should be ignored.

        Parameters:
            exceptions (list): List of exceptions to ignore.
        """
        if isinstance(exceptions, list):
            self.ignored_exceptions = exceptions

    def filter_keys(self, keys: list[str]) -> None:
        """
        Set keys to be filtered out from the error reports.

        Parameters:
            keys (list): List of keys to filter out.
        """
        if isinstance(keys, list):
            self.filtered_keys = keys

    def set_proxy(self, host: str, port: int) -> None:
        """
        Set a proxy for the sender.

        Parameters:
            host (str): Proxy host.
            port (int): Proxy port.
        """
        self.proxy = {"host": host, "port": port}

    def on_before_send(self, callback: BeforeSendCallback) -> None:
        """
        Set a callback to be executed before sending a report.

        Parameters:
            callback (callable): Callback function to be executed.
        """
        if callable(callback):
            self.before_send_callback = callback

    def on_grouping_key(self, callback: GroupingKeyCallback) -> None:
        """
        Set a callback to customize the grouping key for the report.

        Parameters:
            callback (callable): Callback function to customize grouping.
        """
        if callable(callback):
            self.grouping_key_callback = callback

    def send_exception(
        self,
        exception: BaseException | None = None,
        exc_info: ExcInfo | tuple[None, None, None] | None = None,
        user_override: UserInfo = None,
        **kwargs: Any,
    ) -> tuple[int, str] | None:
        """
        Send an exception report to Raygun.

        Parameters:
            exception (Exception, optional): An exception instance to report.
            exc_info (tuple, optional): A 3-tuple containing exception type, exception instance, and traceback.
            user_override (dict or str, optional): Information about the affected user. If not provided, 'self.user' will be used.
            tags (list, optional): A list of tags relating to the current context which you can define.
            userCustomData (dict, optional): A dictionary containing custom key-values also of your choosing.
            httpRequest (dict, optional): HTTP Request data that you wish to include with the report.

        Returns:
            The result of the post request, typically indicating the success or failure of the exception report transmission.
        """
        options = {
            "transmitLocalVariables": self.transmit_local_variables,
            "transmitGlobalVariables": self.transmit_global_variables,
            "enforce_payload_size_limit": self.enforce_payload_size_limit,
            "log_payload_size_limit_breaches": self.log_payload_size_limit_breaches,
        }
        (
            tags,
            custom_data,
            http_request,
            extra_environment_data,
            custom_message,
            fallback_error,
        ) = self._parse_args(kwargs)

        exc_type, exc_value, exc_traceback = exc_info or sys.exc_info()
        errorMessage: (
            raygunmsgs.RaygunErrorMessage | raygunmsgs.RaygunLoggerFallbackErrorMessage
        )
        if fallback_error and not (exc_type or exception):
            errorMessage = fallback_error
        else:
            errorMessage = self._create_error_message(
                exception, exc_type, exc_value, exc_traceback, options, custom_message
            )

        message = self._create_message(
            errorMessage,
            tags,
            custom_data,
            http_request,
            extra_environment_data,
            user_override,
        )
        transformed = self._transform_message(message)

        if transformed is not None:
            return self._post(transformed)
        return None

    def _create_error_message(
        self,
        exception: BaseException | None,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
        options: dict[str, Any],
        custom_message: str | None = None,
    ) -> raygunmsgs.RaygunErrorMessage:
        if exception:
            return raygunmsgs.RaygunErrorMessage(
                type(exception),
                exception,
                exception.__traceback__,
                options,
                custom_message,
            )
        else:
            return raygunmsgs.RaygunErrorMessage(
                exc_type, exc_value, exc_traceback, options, custom_message
            )

    def _parse_args(
        self, kwargs: dict[str, Any]
    ) -> tuple[
        Tags,
        CustomData,
        HttpRequest,
        dict[str, Any] | None,
        str | None,
        raygunmsgs.RaygunLoggerFallbackErrorMessage | None,
    ]:
        tags: Tags = kwargs.get("tags")
        custom_data: CustomData = kwargs.get("userCustomData")
        extra_environment_data: dict[str, Any] | None = kwargs.get(
            "extra_environment_data"
        )
        custom_message: str | None = kwargs.get("custom_message")
        fallback_error: raygunmsgs.RaygunLoggerFallbackErrorMessage | None = kwargs.get(
            "fallback_error"
        )

        http_request: HttpRequest = None
        if "httpRequest" in kwargs:
            http_request = kwargs["httpRequest"]
        elif "request" in kwargs:
            http_request = kwargs["request"]

        return (
            tags,
            custom_data,
            http_request,
            extra_environment_data,
            custom_message,
            fallback_error,
        )

    def _create_message(
        self,
        raygunExceptionMessage: raygunmsgs.RaygunErrorMessage
        | raygunmsgs.RaygunLoggerFallbackErrorMessage,
        tags: Tags,
        user_custom_data: CustomData,
        http_request: HttpRequest,
        extra_environment_data: dict[str, Any] | None,
        user_override: UserInfo = None,
    ) -> raygunmsgs.RaygunMessage:
        options = {
            "transmit_environment_variables": self.transmit_environment_variables
        }
        return (
            raygunmsgs.RaygunMessageBuilder(options)
            .new()
            .set_machine_name(socket.gethostname())
            .set_version(self.userversion)
            .set_client_details()
            .set_exception_details(raygunExceptionMessage)
            .set_environment_details(extra_environment_data)
            .set_tags(self.process_tags)
            .set_tags(tags)
            .set_customdata(self.process_custom_data)
            .set_customdata(user_custom_data)
            .set_request_details(http_request)
            .set_user(user_override if user_override else self.user)
            .build()
        )

    def _transform_message(
        self, message: raygunmsgs.RaygunMessage
    ) -> raygunmsgs.RaygunMessage | None:
        message = message.copy()
        result = utilities.ignore_exceptions(self.ignored_exceptions, message)

        if result is not None:
            details = result.get_details()
            details = utilities.filter_keys(self.filtered_keys, details)
            details["groupingKey"] = utilities.execute_grouping_key(
                self.grouping_key_callback, result
            )
            result.set_details(details)

        if self.before_send_callback is not None and result is not None:
            mutated_payload = self.before_send_callback(result.get_details())

            if mutated_payload is not None:
                result.set_details(mutated_payload)
            else:
                return None

        return result

    def _post(self, raygunMessage: raygunmsgs.RaygunMessage) -> tuple[int, str]:
        raygunMessage = raygunMessage.copy()
        options = {
            "enforce_payload_size_limit": self.enforce_payload_size_limit,
            "log_payload_size_limit_breaches": self.log_payload_size_limit_breaches,
        }

        error = raygunMessage.get_error()
        if (
            isinstance(error, raygunmsgs.RaygunErrorMessage)
            and "enforce_payload_size_limit" in options
            and options["enforce_payload_size_limit"] is True
        ):
            error.check_and_modify_payload_size(options)
            raygunMessage.set_error(error)

        json = jsonpickle.encode(raygunMessage, unpicklable=False)

        try:
            headers = {
                "X-ApiKey": self.api_key,
                "Content-Type": "application/json",
                "User-Agent": "raygun4py",
            }

            response = requests.post(
                self.endpointprotocol + self.endpointhost + self.endpointpath,
                headers=headers,
                data=json,
                timeout=self.http_timeout,
            )
        except Exception as e:
            self.log.error(e)
            return 400, "Exception: Could not send"
        return response.status_code, response.text


class RaygunHandler(logging.Handler):
    """
    Logging handler for sending error logs to Raygun.

    Attributes:
        sender (RaygunSender): The RaygunSender to use for sending.
        version (str): Version for the RaygunSender.
    """

    sender: RaygunSender
    version: str | None

    def __init__(
        self,
        api_key: str | None = None,
        version: str | None = None,
        level: int = logging.ERROR,
        sender: RaygunSender | None = None,
    ) -> None:
        """
        Initialize a RaygunHandler for logging.

        For constructing instances using an existing RaygunSender, it is recommended to use the `RaygunHandler.from_sender()` class method.

        Parameters:
            api_key (str): The API key for Raygun.
            version (str, optional): Version for the RaygunSender. Defaults to None.
            level (int, optional): Logging level. Defaults to logging.ERROR.
        """
        super().__init__(level)
        if api_key:
            self.sender = RaygunSender(api_key)
            if version:
                self.sender.set_version(version)
            self.version = version
        elif sender:
            self.sender = sender
            self.version = sender.userversion
        else:
            raise ValueError("Either 'api_key' or 'sender' must be provided.")

    @classmethod
    def from_sender(
        cls, sender: RaygunSender, level: int = logging.ERROR
    ) -> RaygunHandler:
        """
        Construct a RaygunHandler instance using an existing RaygunSender object.

        Parameters:
            sender (RaygunSender): The RaygunSender to use for sending.
            level (int, optional): Logging level. Defaults to logging.ERROR.

        Returns:
            RaygunHandler: A new instance of RaygunHandler.
        """
        if not isinstance(sender, RaygunSender):
            raise TypeError(
                f"Expected sender of type RaygunSender, but got {type(sender).__name__}"
            )
        return cls(level=level, sender=sender)

    def emit(self, record: logging.LogRecord) -> None:
        # Use log level as a tag
        tag = self.get_tag_from_levelname(record.levelname)
        tags = [tag] if tag else []

        if record.exc_info:
            # exc_info was provided, so send it
            userCustomData = {
                "Logger Name": record.name,
            }
            self.sender.send_exception(
                exc_info=record.exc_info,
                userCustomData=userCustomData,
                tags=tags,
                custom_message=record.getMessage(),
            )
        else:
            # Provide a fallback error to use if exc_info cannot be obtained from sys
            fallback_error = raygunmsgs.RaygunLoggerFallbackErrorMessage(
                record.name,
                record.getMessage(),
                record.filename,
                record.funcName,
                record.lineno,
            )
            self.sender.send_exception(tags=tags, fallback_error=fallback_error)

    @staticmethod
    def get_tag_from_levelname(levelname: str) -> str | None:
        tag_map = {
            "DEBUG": "Debug Log",
            "INFO": "Info Log",
            "WARNING": "Warning Log",
            "ERROR": "Error Log",
            "CRITICAL": "Critical Log",
        }
        return tag_map.get(levelname)



from __future__ import annotations

from typing import Any


def build_wsgi_compliant_request(request: dict[str, Any]) -> dict[str, Any] | None:
    if not request:
        return None

    rg_request: dict[str, Any] | None = None
    # start with WSGI environ variables, then overlay the specific, expected
    # httpRequest keys for RG API compat
    # https://www.python.org/dev/peps/pep-3333/#environ-variables

    http_host = (
        request.get("hostName")
        or request.get("HTTP_HOST")
        or "{}:{}".format(request.get("SERVER_NAME"), request.get("SERVER_PORT"))
    )

    if http_host is not None:
        http_host = http_host.replace(" ", "")

    try:
        http_form = getattr(request, "form", None) or request.get("form")

        # fallback to wsgi.input
        if http_form is None and "wsgi.input" in request:
            # we can assume WSGI keys inside this block
            content_length = int(request.get("CONTENT_LENGTH", 0))
            if content_length:
                http_form = request["wsgi.input"].read(content_length)

        rg_request = {
            "httpMethod": (request.get("httpMethod") or request.get("REQUEST_METHOD")),
            "url": (request.get("url") or request.get("PATH_INFO")),
            "hostName": http_host,
            "iPAddress": (
                request.get("ipAddress")
                or request.get("iPAddress")
                or request.get("REMOTE_ADDR")
            ),
            "queryString": (request.get("queryString") or request.get("QUERY_STRING")),
            "headers": {},  # see below
            "form": http_form,
            "rawData": request.get("rawData"),
        }
    except Exception:
        pass

    if rg_request is None:
        return None

    # Header processing
    _headers = request.get("headers")
    if _headers is None:
        # manually try to build up headers given known WSGI keys
        _headers = {
            "Content-Type": request.get("CONTENT_TYPE"),
            "Content-Length": request.get("CONTENT_LENGTH"),
        }

        for key, value in request.items():
            if key.startswith("HTTP_"):
                # 'HTTP_REFERER' => 'Referer'
                new_key = http_environ_var_to_header_key(key)
                _headers[new_key] = value

    rg_request["headers"] = _headers

    return rg_request


def http_environ_var_to_header_key(key: str) -> str:
    "turns HTTP_REFERER like keys into Referer"
    parts = key.split("_")
    if parts[0] == "HTTP":
        parts.pop(0)

    return "-".join([x.title() for x in parts])



import django
from django.conf import settings

from raygun4py import raygunprovider

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object


class Provider(MiddlewareMixin):
    def __init__(self, get_response=None):
        self.get_response = get_response
        config = getattr(settings, "RAYGUN4PY_CONFIG", {})
        apiKey = getattr(settings, "RAYGUN4PY_API_KEY", config.get("api_key", None))

        self.sender = raygunprovider.RaygunSender(apiKey, config=config)

    def process_exception(self, request, exception):
        raygun_request = self._mapRequest(request)
        env = self._get_django_environment()

        self.sender.send_exception(
            exception=exception, request=raygun_request, extra_environment_data=env
        )

    def _mapRequest(self, request):
        headers = request.META.items()
        _headers = dict()
        for k, v in headers:
            if not k.startswith("wsgi"):
                _headers[k] = v

        return {
            "hostName": request.get_host(),
            "url": request.path,
            "httpMethod": request.method,
            "ipAddress": request.META.get("REMOTE_ADDR", "?"),
            "queryString": dict((key, request.GET[key]) for key in request.GET),
            "form": dict((key, request.POST[key]) for key in request.POST),
            "headers": _headers,
            "rawData": (
                request.data
                if hasattr(request, "data")
                else getattr(request, "raw_post_data", {})
            ),
        }

    def _get_django_environment(self):
        return {"frameworkVersion": django.get_version()}



from __future__ import annotations

import re
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from raygun4py import raygunmsgs


def ignore_exceptions(
    ignored_exceptions: list[type[Exception]],
    message: raygunmsgs.RaygunMessage,
) -> raygunmsgs.RaygunMessage | None:
    error = message.get_error()
    if error is None:
        return message
    classname = error.get_classname()
    if classname and classname in [e.__name__ for e in ignored_exceptions]:
        return None

    return message


def filter_keys(filtered_keys: list[str], obj: dict[str, Any]) -> dict[str, Any]:
    """
    Filter keys from a dictionary.

    Parameters:
        filtered_keys (list): A list of keys to filter.
        obj (dict): The dictionary to filter.

    Returns:
        dict: The filtered dictionary.
    """
    iteration_target = dict(obj)

    for key in iter(iteration_target.keys()):
        if isinstance(iteration_target[key], dict):
            iteration_target[key] = filter_keys(filtered_keys, iteration_target[key])
        else:
            for filter_key in filtered_keys:
                # Wildcard matching
                if "*" in filter_key:
                    filter_key_sanitised = filter_key.replace("*", "")
                    # `*foo*` Match any key that contains the filter key
                    if filter_key.startswith("*") and filter_key.endswith("*"):
                        if filter_key_sanitised in key:
                            iteration_target[key] = "<filtered>"
                    # `*foo` Match any key that ends with the filter key
                    elif filter_key.startswith("*") and key.endswith(
                        filter_key_sanitised
                    ):
                        iteration_target[key] = "<filtered>"
                    # `foo*` Match any key that starts with the filter key
                    elif filter_key.endswith("*") and key.startswith(
                        filter_key_sanitised
                    ):
                        iteration_target[key] = "<filtered>"
                # Exact matching
                elif filter_key == key:
                    iteration_target[key] = "<filtered>"

    return iteration_target


def execute_grouping_key(
    grouping_key_callback: Callable[[raygunmsgs.RaygunMessage], str] | None,
    message: raygunmsgs.RaygunMessage,
) -> str | None:
    if grouping_key_callback is not None:
        grouping_key = grouping_key_callback(message)

        if (
            grouping_key is not None
            and isinstance(grouping_key, str)
            and 0 < len(grouping_key) <= 100
        ):
            return grouping_key

    return None


def camelcase_to_snakecase(key: str) -> str:
    "Turns camelCaseStrings into snake_case_strings."
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", key)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def snakecase_dict(d: dict[str, Any]) -> dict[str, Any]:
    return dict([(camelcase_to_snakecase(k), v) for k, v in d.items()])



from __future__ import annotations

import inspect
import logging
import os
import sys
from types import FrameType, TracebackType
from typing import Any

import jsonpickle

from raygun4py import __version__

try:
    import multiprocessing

    USE_MULTIPROCESSING = True
except ImportError:
    USE_MULTIPROCESSING = False

import platform
from datetime import datetime, timezone

from raygun4py import http_utilities


class RaygunMessageBuilder:
    """Builder class for constructing RaygunMessage objects."""

    def __init__(self, options: dict[str, Any]) -> None:
        self.raygunMessage = RaygunMessage()
        self.options = options

    def new(self) -> RaygunMessageBuilder:
        return RaygunMessageBuilder(self.options)

    def build(self) -> RaygunMessage:
        return self.raygunMessage

    def set_machine_name(self, name: str) -> RaygunMessageBuilder:
        self.raygunMessage.details["machineName"] = name
        return self

    def set_environment_details(
        self, extra_environment_data: dict[str, Any] | None
    ) -> RaygunMessageBuilder:
        self.raygunMessage.details["environment"] = {
            "environmentVariables": dict(os.environ),
            "runtimeLocation": sys.executable,
            "runtimeVersion": "Python " + sys.version,
        }

        if self.options.get("transmit_environment_variables", True) is False:
            self.raygunMessage.details["environment"]["environmentVariables"] = None

        # Wrap these so we gracefully fail if we cannot access the system details for any reason
        try:
            self.raygunMessage.details["environment"]["processorCount"] = (
                multiprocessing.cpu_count() if USE_MULTIPROCESSING else "n/a"
            )
        except Exception:  # pragma: no cover
            pass

        try:
            self.raygunMessage.details["environment"]["architecture"] = (
                platform.architecture()[0]
            )
        except Exception:  # pragma: no cover
            pass

        try:
            self.raygunMessage.details["environment"]["cpu"] = platform.processor()
        except Exception:  # pragma: no cover
            pass

        try:
            self.raygunMessage.details["environment"]["oSVersion"] = "%s %s" % (
                platform.system(),
                platform.release(),
            )
        except Exception:  # pragma: no cover
            pass

        if extra_environment_data is not None:
            merged = extra_environment_data.copy()
            merged.update(self.raygunMessage.details["environment"])
            self.raygunMessage.details["environment"] = merged

        return self

    def set_exception_details(
        self,
        raygunExceptionMessage: RaygunErrorMessage | RaygunLoggerFallbackErrorMessage,
    ) -> RaygunMessageBuilder:
        self.raygunMessage.details["error"] = raygunExceptionMessage
        return self

    def set_client_details(self) -> RaygunMessageBuilder:
        self.raygunMessage.details["client"] = {
            "name": "raygun4py",
            "version": __version__,
            "clientUrl": "https://github.com/MindscapeHQ/raygun4py",
        }
        return self

    def set_customdata(
        self, user_custom_data: dict[str, Any] | None
    ) -> RaygunMessageBuilder:
        if type(user_custom_data) is dict:
            if not self.raygunMessage.details.get("userCustomData"):
                self.raygunMessage.details["userCustomData"] = dict()
            self.raygunMessage.details["userCustomData"].update(user_custom_data)
        return self

    def set_tags(self, tags: list[str] | None) -> RaygunMessageBuilder:
        if type(tags) is list:
            if not self.raygunMessage.details.get("tags"):
                self.raygunMessage.details["tags"] = []
            self.raygunMessage.details["tags"] += tags
        return self

    def set_request_details(
        self, request: dict[str, Any] | None
    ) -> RaygunMessageBuilder:
        if not request:
            return self

        rg_request_details = http_utilities.build_wsgi_compliant_request(request)
        self.raygunMessage.details["request"] = rg_request_details

        return self

    def set_version(self, version: str) -> RaygunMessageBuilder:
        self.raygunMessage.details["version"] = version
        return self

    def set_user(self, user: dict[str, Any] | str | None) -> RaygunMessageBuilder:
        if user is not None:
            self.raygunMessage.details["user"] = user
        return self


class RaygunMessage:
    """Represents a message to be sent to Raygun."""

    def __init__(self) -> None:
        self.occurredOn: datetime = datetime.now(timezone.utc)
        self.details: dict[str, Any] = {}

    def __copy__(self) -> RaygunMessage:
        new_instance = RaygunMessage()
        new_instance.details = self.details.copy()
        new_instance.occurredOn = self.occurredOn
        return new_instance

    def copy(self) -> RaygunMessage:
        return self.__copy__()

    def get_error(self) -> RaygunErrorMessage | RaygunLoggerFallbackErrorMessage | None:
        return self.details.get("error")

    def get_details(self) -> dict[str, Any]:
        return self.details

    def set_details(self, details: dict[str, Any]) -> None:
        self.details = details

    def set_error(
        self, error: RaygunErrorMessage | RaygunLoggerFallbackErrorMessage
    ) -> None:
        self.details["error"] = error


class RaygunErrorMessage:
    """Represents an error message with stack trace information."""

    log: logging.Logger = logging.getLogger(__name__)

    className: str | None
    message: str
    stackTrace: list[dict[str, Any]]
    globalVariables: dict[str, Any] | None
    data: str
    innerError: RaygunErrorMessage | None

    def __init__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        exc_traceback: TracebackType | None = None,
        options: dict[str, Any] | None = None,
        custom_message: str | None = None,
    ) -> None:
        self.className = exc_type.__name__ if exc_type is not None else None
        self.message = (
            custom_message or "%s: %s" % (exc_type.__name__, exc_value)
            if exc_type is not None
            else "RaygunSender.send_exception() called outside of except block"
        )
        self.stackTrace = []
        self.globalVariables = None
        self.innerError = None

        try:
            frames = inspect.getinnerframes(exc_traceback)  # type: ignore[arg-type]

            if frames:
                for frame in frames:
                    localVariables: dict[str, str] | None = None
                    if (
                        options is not None
                        and "transmitLocalVariables" in options
                        and options["transmitLocalVariables"] is True
                    ):
                        localVariables = self._get_locals(frame[0])

                    self.stackTrace.append(
                        {
                            "lineNumber": frame[2],
                            "className": frame[3],
                            "fileName": frame[1],
                            "methodName": frame[4][0] if frame[4] is not None else None,
                            "localVariables": localVariables,
                        }
                    )
                if (
                    options is not None
                    and "transmitGlobalVariables" in options
                    and options["transmitGlobalVariables"] is True
                    and len(frames) > 0
                ):
                    self.globalVariables = frames[-1][0].f_globals.copy()
        except Exception:
            pass
        finally:
            if "frames" in locals():
                del frames

        self.data = ""

        if isinstance(exc_value, Exception):
            nestedException: BaseException | None = None

            if exc_value.__cause__:
                nestedException = exc_value.__cause__
            elif exc_value.__context__:
                nestedException = exc_value.__context__

            if nestedException is not None:
                self.innerError = RaygunErrorMessage(
                    type(nestedException),
                    nestedException,
                    nestedException.__traceback__,
                    options,
                )

        try:
            jsonpickle.encode(self, unpicklable=False)
        except Exception:
            if self.globalVariables:
                self.globalVariables = None

                try:
                    jsonpickle.encode(self, unpicklable=False)
                except Exception:
                    for stack_entry in self.stackTrace:
                        if "localVariables" in stack_entry:
                            stack_entry["localVariables"] = None

    def check_and_modify_payload_size(
        self, options: dict[str, Any], max_size_kb: int = 128
    ) -> None:
        payload = jsonpickle.encode(self, unpicklable=False)

        while len(payload.encode("utf-8")) > max_size_kb * 1024:
            if not self._remove_largest_variable(
                options
            ):  # Failed to remove either local or global variable
                self.log.warning(
                    "Raygun4Py: Unable to reduce size of payload below 128kb, error will be discarded by Raygun ingestion API"
                )
                break

            payload = jsonpickle.encode(self, unpicklable=False)

    def _remove_largest_variable(self, options: dict[str, Any]) -> bool:
        largest_global_var: str | None = None
        largest_global_size: int = 0

        largest_local_var: str | None = None
        largest_local_size: int = 0
        largest_local_frame: dict[str, Any] | None = None

        # Find the largest global variable
        if self.globalVariables:
            for var, value in self.globalVariables.items():
                size = sys.getsizeof(value)
                if size > largest_global_size:
                    largest_global_size = size
                    largest_global_var = var

        # Find the largest local variable across all frames
        for frame in self.stackTrace:
            if "localVariables" in frame and frame["localVariables"]:
                for var, value in frame["localVariables"].items():
                    size = sys.getsizeof(value)
                    if size > largest_local_size:
                        largest_local_size = size
                        largest_local_var = var
                        largest_local_frame = frame

        if largest_global_size == 0 and largest_local_size == 0:
            return False

        # Decide which one to remove: the largest global or the largest local variable
        if largest_global_size >= largest_local_size and largest_global_var is not None:
            if (
                options is not None
                and "log_payload_size_limit_breaches" in options
                and options["log_payload_size_limit_breaches"] is True
            ):
                self.log.warning(
                    f"Raygun4Py: Removing global variable {largest_global_var} due to payload size limit"
                )

            if self.globalVariables is not None:
                self.globalVariables[largest_global_var] = "Removed"
        elif largest_local_var is not None and largest_local_frame is not None:
            if (
                options is not None
                and "log_payload_size_limit_breaches" in options
                and options["log_payload_size_limit_breaches"] is True
            ):
                self.log.warning(
                    f"Raygun4Py: Removing local variable {largest_local_var} due to payload size limit"
                )

            largest_local_frame["localVariables"][largest_local_var] = "Removed"

        return True

    def get_classname(self) -> str | None:
        return self.className

    def _get_locals(self, frame: FrameType) -> dict[str, str]:
        result: dict[str, str] = {}
        localVars = getattr(frame, "f_locals", {})

        if "__traceback_hide__" not in localVars:
            for key in localVars:
                try:
                    # Note that str() *can* fail; thus protect against it as much as we can.
                    result[key] = str(localVars[key])
                except Exception as e:
                    try:
                        r = repr(localVars[key])
                    except Exception as re:
                        r = "Couldn't convert to repr due to {0}".format(re)
                    result[key] = (
                        "!!! Couldn't convert {0!r} (repr: {1}) due to {2!r} !!!".format(
                            key, r, e
                        )
                    )
        return result


class RaygunLoggerFallbackErrorMessage:
    """Fallback error message for use when logging without an exception context."""

    className: str
    message: str
    stackTrace: list[dict[str, Any]]
    globalVariables: None
    data: str

    def __init__(
        self,
        name: str,
        message: str,
        filename: str,
        funcName: str | None,
        lineno: int,
    ) -> None:
        self.className = "Logger (" + name + ")"
        self.message = message
        # Create a single stackTrace entry using logger data
        self.stackTrace = [
            {
                "lineNumber": lineno,
                "className": self.className,
                "fileName": filename,
                "methodName": funcName or "UnknownMethod",
                "localVariables": None,  # We don't have access to local variables
            }
        ]
        self.globalVariables = None  # We don't have access to global variables
        self.data = ""

        try:
            jsonpickle.encode(self, unpicklable=False)
        except Exception:
            for frame in self.stackTrace:
                if "localVariables" in frame:
                    frame["localVariables"] = None

    def get_classname(self) -> str:
        return self.className



# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) hatchling can read it for dynamic versioning (pyproject.toml)
# 3) we can import it into your module
__version__ = "7.0.0"
